#!/usr/bin/env python
#(c)2020, Anton Karneliuk

# Modules
import grpc
from pygnmi.spec.gnmi_pb2_grpc import gNMIStub
from pygnmi.spec.gnmi_pb2 import CapabilityRequest, GetRequest, SetRequest
import re
import sys
import json
import logging


# Own modules
from pygnmi.path_generator import gnmi_path_generator


# Classes
class gNMIclient(object):
    """
    This class instantiates the object, which interacts with the network elements over gNMI.
    """
    def __init__(self, target, username=None, password=None, to_print=False, insecure=False, path_cert=None):
        """
        Initializing the object
        """
        self.__metadata = [('username', username), ('password', password)]
        self.__target = target
        self.__capabilities = None
        self.__to_print = to_print
        self.__insecure = insecure
        self.__path_cert = path_cert

    
    def __enter__(self):
        """
        Building the connectivity towards network element over gNMI
        """
        if self.__insecure:
            self.__channel = grpc.insecure_channel(f'{self.__target[0]}:{self.__target[1]}', self.__metadata)
            grpc.channel_ready_future(self.__channel).result(timeout=5)
            self.__stub = gNMIStub(self.__channel)

        else:
            if self.__path_cert:
                try:
                    with open(self.__path_cert, 'rb') as f:
                        cert = grpc.ssl_channel_credentials(f.read())

                except:
                    logging.error('The SSL certificate cannot be opened.')
                    sys.exit(10)

            self.__channel = grpc.secure_channel(f'{self.__target[0]}:{self.__target[1]}', cert)
            grpc.channel_ready_future(self.__channel).result(timeout=5)
            self.__stub = gNMIStub(self.__channel)

        return self


    def capabilities(self):
        """
        Collecting capabilities
        """
        logging.info(f'Collecting Capabilities...')

        try:
            gnmi_message_request = CapabilityRequest()
            gnmi_message_response = self.__stub.Capabilities(gnmi_message_request, metadata=self.__metadata)

            if self.__to_print:
                print(gnmi_message_response)

            if gnmi_message_response:
                response = {}
                
                if gnmi_message_response.supported_models:
                    response.update({'supported_models': []})

                    for ree in gnmi_message_response.supported_models:
                        response['supported_models'].append({'name': ree.name, 'organization': ree.organization, 'version': ree.version})

                if gnmi_message_response.supported_encodings:
                    response.update({'supported_encodings': []})

                    for ree in gnmi_message_response.supported_encodings:
                        response['supported_encodings'].append(ree)

                if gnmi_message_response.gNMI_version:
                    response.update({'gnmi_version': gnmi_message_response.gNMI_version})

            logging.info(f'Collection of Capabilities is successfull')

            self.__capabilities = response
            return response

        except:
            logging.error(f'Collection of Capabilities is failed.')


    def get(self, path):
        """
        Collecting path information
        """
        logging.info(f'Collecting info from requested paths (Get opertaion)...')

        protobuf_paths = [gnmi_path_generator(pe) for pe in path]

        if self.__capabilities and 'supported_encodings' in self.__capabilities:
            if 0 in self.__capabilities['supported_encodings']:
                enc = 0
            elif 4 in self.__capabilities['supported_encodings']:
                enc = 4

        else:
            enc = 0

        gnmi_message_request = GetRequest(path=protobuf_paths, type=0, encoding=enc)
        gnmi_message_response = self.__stub.Get(gnmi_message_request, metadata=self.__metadata)

        if self.__to_print:
            print(gnmi_message_response)


    def __exit__(self, type, value, traceback):
        self.__channel.close()


msg = {
    'unknown_arg': 'There is no such argument. Run \'--help\' for details.',
    'help': 'The following keys are availble:\n\n  -u (--user):      Provide the username to connect to the network element\n  -p (--pass):      Provide the password to connect to the network element\n  -h (--help):      Provide the help on the available keys\n  -t (--target):    Provide the list of endpoint in \'host:port\' format\n  -o (--operation): Prvoide the type of gNMI request\n  -c (--cert):      Provide the path towards the certificate\n  --insecure:       Define whether gRPC channel is encrypted or not\n  --gnmi-path:      Provide path to the resource at the network function\n  --print:          Define whether Protobuf messages shall be printed in the STDOUT',
    'bad_host': 'The host address is malformed. It shall be provided in a format \'ip_address:grpc_port\'.',
    'not_defined_user': 'The username is not defined. The execution is terminated.',
    'not_defined_pass': 'The password is not defined.',
    'not_defined_target': 'There are no hosts provided. The execution is terminated.',
    'not_enough_arg': 'There are not enough arguments.',
    'wrong_data': 'The argument is provided in thw wrong type (e.g. string instead of integer).',
    'not_allowed_op': 'The request gNMI operation type is now allowed.',
    'not_defined_op': 'The gNMI operation type is not defined.',
    'not_defined_path': 'The gNMI path is ot defined with Get or Set operation.'
}


#!/usr/bin/env python

# Modules
from pygnmi.spec.gnmi_pb2 import *
import re

# User-defined functions
def gnmi_path_generator(path_in_question):
    gnmi_path = Path()
    keys = []

    # Subtracting all the keys from the elements and storing them separately
    while re.match('.*?\[.+?=.+?\].*?', path_in_question):
        temp_key, temp_value = re.sub('.*?\[(.+?)\].*?', '\g<1>', path_in_question).split('=')
        keys.append({temp_key: temp_value})
        path_in_question = re.sub('(.*?\[).+?(\].*?)', f'\g<1>{len(keys) - 1}\g<2>', path_in_question)

    path_elements = path_in_question.split('/')

    for pe_entry in path_elements:
        if not re.match('.+?:.+?', pe_entry) and len(path_elements) == 1:
            sys.exit(f'You haven\'t specified either YANG module or the top-level container in \'{pe_entry}\'.')

        elif re.match('.+?:.+?', pe_entry):
            gnmi_path.origin = pe_entry.split(':')[0]
            gnmi_path.elem.add(name=pe_entry.split(':')[1])

        elif re.match('.+?\[\d+?\].*?', pe_entry):
            key_id = int(re.sub('.+?\[(\d+?)\].*?', '\g<1>', pe_entry))
            gnmi_path.elem.add(name=pe_entry.split('[')[0], key=keys[key_id])

        else:
            gnmi_path.elem.add(name=pe_entry)

    return gnmi_path


from distutils.core import setup

with open('README.rst') as fh:
    long_description = fh.read()

setup(
  name = 'pygnmi',
  packages = ['pygnmi', 'pygnmi.spec', 'pygnmi.artefacts'],
  version = '0.1.7',
  license='bsd-3-clause',
  description = 'This repository contains pure Python implementation of the gNMI client to interact with the network functions.',
  long_description = long_description,
  long_description_content_type = 'text/x-rst',
  author = 'Anton Karneliuk',
  author_email = 'anton@karneliuk.com',
  url = 'https://github.com/akarneliuk/pygnmi',
  download_url = 'https://github.com/akarneliuk/pygnmi/archive/v0.1.7.tar.gz',
  keywords = ['gnmi', 'automation', 'grpc', 'network'], 
  install_requires=[
          'grpcio',
          'grpcio-tools'
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Intended Audience :: System Administrators',
    'Intended Audience :: Telecommunications Industry',
    'Operating System :: OS Independent',
    'Topic :: System :: Networking',
    'License :: OSI Approved :: BSD License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
  ],
)